#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

import logging

import dpath.util
import pytest
from airbyte_cdk.sources.streams.files import FilesSpec
from pydantic import BaseModel, Field

logger = logging.getLogger("airbyte")


class MockBrokenSpec(FilesSpec, BaseModel):
    class Config:
        title = "Mock Spec"


class MockSpec(FilesSpec, BaseModel):
    class Config:
        title = "Mock Spec"

    class MockProvider(BaseModel):
        class Config:
            title = "Mock Provider"

    provider: MockProvider = Field(...)


class TestFilesSpec:
    def test_post_processing_no_provider(self):
        with pytest.raises(RuntimeError):
            MockBrokenSpec.schema()

    def test_post_processing(self):
        out_schema = MockSpec.schema()
        assert len([x for x in dpath.util.search(out_schema, "**/anyOf")]) == 0
        assert len([x for x in dpath.util.search(out_schema, "**/oneOf")]) > 0
        assert len([x for x in dpath.util.search(out_schema, "**/$ref")]) == 0

    def test_without_post_processing(self):
        # we should prove that the above assertions would NOT hold without the added post-processing

        delattr(FilesSpec, "schema")  # this removes our method override

        # and now we create a class based on FilesSpec without the schema() override
        class MockSpecWithoutPostProcessing(FilesSpec, BaseModel):
            class Config:
                title = "Mock Spec"

            class MockProvider(BaseModel):
                class Config:
                    title = "Mock Provider"

            provider: MockProvider = Field(...)

        out_schema = MockSpecWithoutPostProcessing.schema()
        assert len([x for x in dpath.util.search(out_schema, "**/anyOf")]) > 0
        assert len([x for x in dpath.util.search(out_schema, "**/oneOf")]) == 0
        assert len([x for x in dpath.util.search(out_schema, "**/$ref")]) > 0



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from abc import ABC, abstractmethod
from typing import Any, Mapping

from airbyte_cdk.sources.declarative.types import Config, StreamSlice, StreamState


class RecordTransformation(ABC):
    """
    Implementations of this class define transformations that can be applied to records of a stream.
    """

    @abstractmethod
    def transform(
        self, record: Mapping[str, Any], config: Config = None, stream_state: StreamState = None, stream_slice: StreamSlice = None
    ) -> Mapping[str, Any]:
        """
        :param record: the input record to be transformed
        :return: the transformed record
        """

    def __eq__(self, other):
        return other.__dict__ == self.__dict__



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

import json
import logging
from typing import Any, List, Mapping

from airbyte_cdk.sources.declarative.checks.connection_checker import ConnectionChecker
from airbyte_cdk.sources.declarative.declarative_source import DeclarativeSource
from airbyte_cdk.sources.declarative.parsers.factory import DeclarativeComponentFactory
from airbyte_cdk.sources.declarative.parsers.yaml_parser import YamlParser
from airbyte_cdk.sources.streams import Stream


class YamlDeclarativeSource(DeclarativeSource):
    def __init__(self, path_to_yaml):
        self.logger = logging.getLogger(f"airbyte.{self.name}")
        self.logger.setLevel(logging.DEBUG)
        self._factory = DeclarativeComponentFactory()
        self._source_config = self._read_and_parse_yaml_file(path_to_yaml)

    @property
    def connection_checker(self) -> ConnectionChecker:
        check = self._source_config["check"]
        if "class_name" not in check:
            check["class_name"] = "airbyte_cdk.sources.declarative.checks.check_stream.CheckStream"
        return self._factory.create_component(check, dict())(source=self)

    def streams(self, config: Mapping[str, Any]) -> List[Stream]:
        stream_configs = self._source_config["streams"]
        for s in stream_configs:
            if "class_name" not in s:
                s["class_name"] = "airbyte_cdk.sources.declarative.declarative_stream.DeclarativeStream"
        return [self._factory.create_component(stream_config, config)() for stream_config in self._source_config["streams"]]

    def _read_and_parse_yaml_file(self, path_to_yaml_file):
        with open(path_to_yaml_file, "r") as f:
            config_content = f.read()
            parsed_config = YamlParser().parse(config_content)
            self.logger.debug(
                "parsed YAML into declarative source",
                extra={"path_to_yaml_file": path_to_yaml_file, "source_name": self.name, "parsed_config": json.dumps(parsed_config)},
            )
            return parsed_config



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from enum import Enum
from typing import Mapping

from airbyte_cdk.sources.declarative.interpolation.jinja import JinjaInterpolation
from airbyte_cdk.sources.declarative.states.state import State


def _get_max(*, name, val, other_state):
    other_val = other_state.get(name)
    if other_val:
        return max(val, other_val)
    else:
        return val


class StateType(Enum):
    STR = str
    INT = int


class DictState(State):
    stream_state_field = "stream_state"

    def __init__(self, initial_mapping: Mapping[str, str] = None, config=None):
        if initial_mapping is None:
            initial_mapping = dict()
        if config is None:
            config = dict()
        self._templates_to_evaluate = initial_mapping
        self._interpolator = JinjaInterpolation()
        self._context = dict()
        self._config = config

    def set_state(self, state):
        self._context[self.stream_state_field] = state

    def update_state(self, **kwargs):
        stream_state = kwargs.get(self.stream_state_field)
        prev_stream_state = self.get_stream_state() or stream_state
        self._context.update(**kwargs)

        self._context[self.stream_state_field] = self._compute_state(prev_stream_state)

    def get_state(self, state_field):
        return self._context.get(state_field, {})

    def get_stream_state(self):
        return self.get_state(self.stream_state_field)

    def _compute_state(self, prev_state):
        updated_state = {
            self._interpolator.eval(name, self._config): self._interpolator.eval(value, self._config, **self._context)
            for name, value in self._templates_to_evaluate.items()
        }
        updated_state = {name: value for name, value in updated_state.items() if value}

        if prev_state:
            next_state = {name: _get_max(name=name, val=value, other_state=prev_state) for name, value in updated_state.items()}
        else:
            next_state = updated_state

        self._context[self.stream_state_field] = next_state
        return next_state



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

import pytest as pytest
from airbyte_cdk.models import SyncMode
from airbyte_cdk.sources.declarative.stream_slicers.list_stream_slicer import ListStreamSlicer


@pytest.mark.parametrize(
    "test_name, slice_values, slice_definition, expected_slices",
    [
        (
            "test_single_element",
            ["customer", "store", "subscription"],
            {"owner_resource": "{{ slice_value }}"},
            [{"owner_resource": "customer"}, {"owner_resource": "store"}, {"owner_resource": "subscription"}],
        ),
    ],
)
def test_list_slicer(test_name, slice_values, slice_definition, expected_slices):
    slicer = ListStreamSlicer(slice_values, slice_definition, config={})
    slices = [s for s in slicer.stream_slices(SyncMode.incremental, stream_state=None)]
    assert slices == expected_slices
